/*
 *  QT_utils.h
 *  video_echo3
 *
 *  Created by David Cairns on 3/9/06.
 *
 */

#ifndef DRC_QT_UTILS
#define DRC_QT_UTILS

#include <Carbon/Carbon.h>
#include <Quicktime/Quicktime.h>


// This data structure contains information about an image
// decompression sequence.  When it is created, it should be
// zeroed-out, so we know that it has not been initialized.
struct ImageDecompressionSequence
{
	bool			initialized;
	ImageSequence	decompressSeq;
};
typedef struct ImageDecompressionSequence ImageDecompressionSequence;


// This data structure is to be used to store information about
// a buffer that holds image data.  The data in the buffer can
// then be used in a call to glDrawPixels() or sent into the
// OpenGL server as a texture.
struct ImageBuffer
{
	Rect			bounds;			// Represents the size of our image.
	unsigned char	*buffer;		// A pointer to the image's pixel data.
	unsigned		elemSize;		// The size of a given element in the buffer.
	unsigned		numElems;		// The number of elements in the array.
	int				bytesPerRow;	// The number of bytes in each row of the image.
};
typedef struct ImageBuffer ImageBuffer;


OSErr SetupImageDecompressionSequence( 
	ImageDecompressionSequence	*seq, 
	ImageBuffer					*imageBuffer, 
	WindowRef					windowRef, 
	GWorldPtr					graphicsWorld, 
	SGChannel					channel, 
	Ptr							dataPtr, 
	long						dataLength );

void DestroyImageSequence( ImageDecompressionSequence *imageSequence );

void InitImageBuffer( 
	Rect		bounds, 
	ImageBuffer	*imageBuffer, 
	unsigned	numComponents, 
	unsigned	elementSize );

OSStatus InitImageBufferWithGWorld( 
	GWorldPtr	*graphicsWorld, 
	Rect		bounds, 
	ImageBuffer *imageBuffer, 
	unsigned	numComponents, 
	unsigned	elementSize );

void DestroyImageBuffer( ImageBuffer *imageBuffer );

void LockImageBuffer( ImageBuffer *theImageBuffer, GWorldPtr graphicsWorld );
void UnlockImageBuffer( ImageBuffer *theImageBuffer, GWorldPtr graphicsWorld );

OSStatus SetupSequenceGrabber( 
	SeqGrabComponent	*seqGrabber, 
	SGChannel			*sgVideoChannel, 
	//WindowRef			windowRef, 
	GWorldPtr			*graphicsWorld, 
	Rect				imageBounds, 
	SGDataUPP			dataProcCallback, 
	long				refcon );

void DestroySequenceGrabber( SeqGrabComponent seq, SGChannel sgChannel );


OSStatus MakeSequenceGrabChannel( 
	SeqGrabComponent seqGrabber, 
	SGChannel *sgVideoChannel, 
	Rect *videoBounds );

#endif // DRC_QT_UTILS
