//
//  main.c
//  Stereo_Video
//
//  Created by David Cairns on 5/8/06.
//

#include <math.h>
#include <ctype.h>

#include <Carbon/Carbon.h>
#include <AGL/agl.h>
#include <OpenGL/gl.h>
#include <Accelerate/Accelerate.h>

#include "Carbon_utils.h"
#include "OpenGL_utils.h"
#include "QT_utils.h"
#include "FPS.h"
#include "WinData.h"
#include "FrameData.h"
#include "AllData.h"
#include "SingleImage.h"
#include "StereoVideo.h"



Rect g_imageBounds = { 0, 0, IMAGE_HEIGHT, IMAGE_WIDTH };
AllData *g_pAllData;



// A function for initializing all of the data in our FrameData structure.
OSStatus SetupFrameData( FrameData *pFrameData );

// Our function for drawing stuff in GL.
void DrawGL();

// Our callback function for the sequence grabber.
static pascal OSErr DataProcFunc( 
	SGChannel	channel, 
	Ptr			data, 
	long		DataLength, 
	long		*offset, 
	long		chRefCon, 
	TimeValue	time, 
	short		writeType, 
	long		refCon );


// Our event-handling functions.
static OSStatus AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon );
static pascal void TimerHandler( EventLoopTimerRef inTimer, void *inUserData );

// This will maintain a reference to our "main.nib" file.
static IBNibRef sNibRef;

FrameData*
FrameData_construct(char *name, WinData *pWinData)
{
	FrameData *pFrameData = (FrameData*)malloc( sizeof( FrameData ) );
	pFrameData->imageSequence.initialized = FALSE;
	strncpy(pFrameData->name, name, FDNameSize);
	pFrameData->pWinData = pWinData;
	
	pFrameData->bupdated = false;
	
	return pFrameData;
	
}


OSStatus
FrameData_SetupSequenceGrabber (FrameData *pFrameData, AllData *pAllData)
{
	OSStatus err = noErr;
	
	printf ("FrameData_SetupSequenceGrabber() [%s]\n", pFrameData->name);
	
	// Set up the default sequence grabber component.
	err = SetupSequenceGrabber( 
		&(pFrameData->seqGrabber), 
		&(pFrameData->sgVideoChannel), 
		&(pFrameData->graphicsWorld), 
		pFrameData->buffer.bounds, 
		NewSGDataUPP( DataProcFunc ), 
		(long)pFrameData );
	if( noErr != err ) {
		printf( "Error trying to set up default sequence grabber component.\n" );
	}

	return err;
}




//--------------------------------------------------------------------------------------------
OSStatus StereoVideo_construct(StereoVideo *pStereoVideo)
{
	
	WinData *pWinData = (WinData*)malloc( sizeof( WinData ));

	SetRect( 
		&(pWinData->winBounds), 
		100, 100, 
		100 + WINDOW_WIDTH, 100 + WINDOW_HEIGHT );
		
		
	// Set up the FrameData structure that we are going to pass around.

	FrameData *pFrameDataR = FrameData_construct("Right", pWinData);
	FrameData *pFrameDataL = FrameData_construct("Left", pWinData);

	
	
	AllData *pAllData = AllData_construct();
	pAllData->pWinData = pWinData;
	pAllData->pFrameDataR = pFrameDataR;
	pAllData->pFrameDataL = pFrameDataL;
	pAllData->pStereoVideo = pStereoVideo;
	g_pAllData = pAllData;
	

	
	// Define the bounds for our window (taking into account the offset).
		
	// Create our OpenGL window.
	OSStatus err = noErr;
	err = MakeOpenGLWindow( 
		&(pWinData->glWindow), 
		pWinData->winBounds, 
		&(pWinData->context) );
	if( noErr != err ) {
		printf( "Error, unable to make OpenGL window.\n" );
		return err;
	}
		
	
	// Set up our FrameData structure.
	err = SetupFrameData( pFrameDataR );
	if( noErr != err ) {
		printf( "Error trying to set up our FrameData structure.\n" );
		return 0;
	}
	
	
	
	// Set up our FrameData structure.
	err = SetupFrameData( pFrameDataL );
	if( noErr != err ) {
		printf( "Error trying to set up our FrameData structure.\n" );
		return 0;
	}

	
	//TODO: initialize SingleImage objects here
	
	//TODO: Select the camera here.
	// We want to use our default parameters.
	SetupOpenGLDefaults( pWinData->winBounds );

	if ( (err=FrameData_SetupSequenceGrabber (pFrameDataR, pAllData)) != noErr)
	{
		return err;
	}
	
//#define NOLEFTFRAME
	
#ifndef NOLEFTFRAME
	if ( (err=FrameData_SetupSequenceGrabber (pFrameDataL, pAllData)) != noErr)
	{
		return err;
	}
#endif
	
			
	// Create and install our application and timer event handlers.
	err = SetupAllEventHandlers( 
		NewEventHandlerUPP( AppEventHandler ), 
		NULL, 
		NULL, 
		(EventLoopTimerUPP)TimerHandler, 
		pWinData->glWindow, 
		sNibRef, 
		(long)pAllData);
	if( noErr != err ) {
		printf( "Error setting up event handlers.\n" );
		return err;
	}
	
	
	AllData_allocateVImageBuffers(g_pAllData, &(pFrameDataR->buffer));
	
	
	/*
	printf("GWORLD: %ld %ld %ld %ld (last two are zero)\n", pFrameDataR->graphicsWorld, pFrameDataL->graphicsWorld,
			pFrameDataR->buffer.buffer, pFrameDataL->buffer.buffer);
	printf("TEXTURE: %d %d\n", pFrameDataR->texture, pFrameDataL->texture);
	 */

	return err;
}


//--------------------------------------------------------------------------------------------
OSStatus SetupFrameData( FrameData *pFrameData )
{
	printf ("SetupFrameData()\n");
	// Set up the rect that our image will be scaled to.
	Rect imageBounds = g_imageBounds;
	
	// Setup our ImageBuffer data structure, so we can access the pixel
	// data for each frame that the sequence grabber will draw into it.
	// This ImageBuffer will contain the frames of video, as they are read.
	OSStatus err;
	err = InitImageBufferWithGWorld( 
		&(pFrameData->graphicsWorld), 
		imageBounds, 
		&(pFrameData->buffer), 
		4, 
		sizeof( unsigned char ) );
	if( noErr != err ) {
		printf( "Error trying to set up an ImageBuffer structure.\n" );
		return err;
	}
	

	// Setup the OpenGL textures for our buffers.
	pFrameData->texture = SetupTextureForImageBuffer( 
		&(pFrameData->buffer), 
		pFrameData->graphicsWorld );
	
	return err;
}



//--------------------------------------------------------------------------------------------

void StereoVideo_displayImage(unsigned texture, ImageBuffer *buffer, int left, int right, int top, int bottom)
{
	
	GLfloat width = buffer->bounds.right - buffer->bounds.left;
	GLfloat height = buffer->bounds.bottom - buffer->bounds.top;
	
	
	glBindTexture( GL_TEXTURE_RECTANGLE_EXT, texture );
	UpdateGLTexture( buffer, 0);
	glBegin( GL_QUADS );
	glTexCoord2f( 0, height );
	glVertex2i( left, top );
	
	glTexCoord2f( width, height );
	glVertex2i( right, top );
	
	glTexCoord2f( width, 0 );
	glVertex2i( right, bottom );
	
	glTexCoord2f( 0, 0 );
	glVertex2i( left, bottom );
	glEnd();
	glEnd();
}


void DrawGL( )
{
		
	AllData *pAllData = g_pAllData;
	
	
	if (pAllData->bPaused)
	  {
		return;
	  }
	
	
	//** Measure frame rate; max is 30fps w/iSight cameras
	if (!g_pAllData->pFrameDataL->bupdated
		|| !g_pAllData->pFrameDataR->bupdated)
	{
		return;
	}
	g_pAllData->pFrameDataL->bupdated=false;
	g_pAllData->pFrameDataR->bupdated=false;
		
	FPS_update(pAllData->pFPS);
	if(pAllData->pFPS->bchanged)
	{
		printf("%.2ffps\n", pAllData->pFPS->rate);
	}

	
	
	
	
	// Reset the color and depth buffers, and the modelview matrix.
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glLoadIdentity();
	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
	

	LockImageBuffer( &(pAllData->pFrameDataR->buffer), pAllData->pFrameDataR->graphicsWorld );
	LockImageBuffer( &(pAllData->pFrameDataL->buffer), pAllData->pFrameDataL->graphicsWorld );


	
	pAllData->pStereoVideo->processFrames(pAllData->pFrameDataL, pAllData->pFrameDataR, pAllData->bAB);
	
	
	pAllData->pStereoVideo->drawFrames(pAllData->bDiagnosticVideos, pAllData->bAB);
		
	
	
	UnlockImageBuffer( &(pAllData->pFrameDataL->buffer),  pAllData->pFrameDataL->graphicsWorld );
	UnlockImageBuffer( &(pAllData->pFrameDataR->buffer),  pAllData->pFrameDataR->graphicsWorld );
	
	
	// Swap the buffers.
	aglSwapBuffers( pAllData->pWinData->context );
	
	
}


//--------------------------------------------------------------------------------------------
static pascal void TimerHandler( EventLoopTimerRef inTimer, void *inUserData )
{
	// This function should get called about 30 times per second.
	AllData *pAllData = (AllData*)inUserData;
	ComponentResult err = noErr;
	
	
	Byte isPaused;
		
		

	// If the SG is paused, don't try to do any processing!

	SGGetPause( pAllData->pFrameDataR->seqGrabber, &isPaused);
	if( seqGrabPause != isPaused )
	{
		// First let's tell our sequence grabber to process a little bit.
		err = SGIdle( pAllData->pFrameDataR->seqGrabber );
		if(	noErr != err )
		{
			printf( "SGIdle raised an error: %i.\n", (int)err );
			// If there was an error, let's just try to restart the sequence grabber.
			SGStop( pAllData->pFrameDataR->seqGrabber );
			SGStartRecord( pAllData->pFrameDataR->seqGrabber );
		}
	}


	SGGetPause( pAllData->pFrameDataL->seqGrabber, &isPaused);
	if( seqGrabPause != isPaused )
	{

		err = SGIdle( pAllData->pFrameDataL->seqGrabber );
		if( noErr != err )
		{
			printf( "SGIdle raised an error: %i.\n", (int)err );
			// If there was an error, let's just try to restart the sequence grabber.
			SGStop( pAllData->pFrameDataL->seqGrabber );
			SGStartRecord( pAllData->pFrameDataL->seqGrabber );
		}
	}




}


//--------------------------------------------------------------------------------------------
static OSStatus AppEventHandler( EventHandlerCallRef inCaller, EventRef inEvent, void* inRefcon )
{
	// Set up our reference constant.
	AllData		*pAllData = (AllData*)inRefcon;
    OSStatus    err = eventNotHandledErr;
	
	//printf ("AppEventHandler\n");
	
	if( NULL == pAllData ) {
		printf( "Error: Invalid pointer to FrameData structure in AppEventHandler.\n" );
		return err;
	}
	
	// Get some info about the event.
	UInt32 eventClass	= GetEventClass( inEvent );
	
	HICommandExtended commandStruct;
				
	switch (eventClass)
	{
		case kCoreEventClass:
			DestroySequenceGrabber( pAllData->pFrameDataR->seqGrabber, pAllData->pFrameDataR->sgVideoChannel );
			DestroySequenceGrabber( pAllData->pFrameDataL->seqGrabber, pAllData->pFrameDataL->sgVideoChannel );

			DestroyImageBuffer( &(pAllData->pFrameDataR->buffer) );
			DestroyImageBuffer( &(pAllData->pFrameDataL->buffer) );

			DestroyImageSequence( &(pAllData->pFrameDataR->imageSequence) );
			DestroyImageSequence( &(pAllData->pFrameDataL->imageSequence) );
		
			err = noErr;
			break;
		case kEventClassCommand:
			GetEventParameter (inEvent, kEventParamDirectObject, 
							   typeHICommand, NULL, sizeof(struct HICommandExtended), 
							   NULL, &commandStruct);
			
			switch (commandStruct.commandID)
			  {
				case 'SWAP':
					printf("Swap cameras\n");
					FrameData *pFrameDataTemp;
					pFrameDataTemp = g_pAllData->pFrameDataL;
					g_pAllData->pFrameDataL=g_pAllData->pFrameDataR;
					g_pAllData->pFrameDataR=pFrameDataTemp;
					char tmp[FDNameSize];
					strncpy(tmp, g_pAllData->pFrameDataR->name, FDNameSize);
					strncpy(g_pAllData->pFrameDataR->name, g_pAllData->pFrameDataL->name, FDNameSize);
					strncpy(g_pAllData->pFrameDataL->name, g_pAllData->pFrameDataR->name, FDNameSize);
					break;
				case 'DIAG':
					printf("Show diagnostic videos\n");
					g_pAllData->bDiagnosticVideos = 1-g_pAllData->bDiagnosticVideos;
					break;
				case 'PAUS':
					printf("Pause\n");
					g_pAllData->bPaused = 1-g_pAllData->bPaused;
					break;
				case 'AB  ':
					g_pAllData->bAB = 1-g_pAllData->bAB;
					printf("AB %d\n", g_pAllData->bAB);
					break;
			  }
				
				break;
			
	}
    
    return err;
}


OSErr
FrameData_update(FrameData *pFrameData, Ptr data, long dataLength)
{
	
	OSErr err = noErr;
		
	// If this is the first time that this function has been called, then set up
	// our image decompression sequence.
	if( FALSE == pFrameData->imageSequence.initialized ) {
		err = SetupImageDecompressionSequence( 
			&(pFrameData->imageSequence), 
			&(pFrameData->buffer), 
			pFrameData->pWinData->glWindow, 
			pFrameData->graphicsWorld, 
			pFrameData->sgVideoChannel, 
			data, 
			dataLength );
		if( noErr != err ) {
			printf( "Error setting up image decompression sequence.\n" );
			return err;
		}
	}
	
	
	// Decompress a frame from our ImageDecompressionSequence into our GWorld.
	CodecFlags ignore;
	err = DecompressSequenceFrameS( 
		pFrameData->imageSequence.decompressSeq,
		data,
		dataLength,
		0,				// no flags going in
		&ignore,		// we don't care about the flags coming out.
		NULL );			// NULL means we want to use asynchronous completion proc.
	if( noErr != err ) {
		printf( "Error decompressing a frame of our image sequence: %i.\n", err );
		return err;
	}


	
	// We update the textures here instead of in DrawGL() because we
	// want to avoid accidentally unlocking any of the buffers we might
	// use before the end of DrawGL().  We only need to update the texture
	// when more data is available
	//moved -- read the above comment again glBindTexture( GL_TEXTURE_RECTANGLE_EXT, pFrameData->texture );
	//moved  UpdateGLTexture( &(pFrameData->buffer), pFrameData->graphicsWorld );

	pFrameData->bupdated=true;
	
	
	return err;
}


//--------------------------------------------------------------------------------------------
static pascal OSErr DataProcFunc( 
	SGChannel	channel, 
	Ptr			data, 
	long		dataLength, 
	long		*offset, 
	long		chRefCon, 
	TimeValue	time, 
	short		writeType, 
	long		refCon )
{
	OSErr err = noErr;
	// Turn the refCon into something we can work with, and make sure it's valid.
	FrameData *pFrameData = (FrameData*)refCon;
	if( NULL == pFrameData ) {
		printf( "Improper refcon sent to SG data processing callback.\n" );
		return -1;
	}


	if ( (err=FrameData_update(pFrameData, data, dataLength)) != noErr )
	{
		printf( "DataProcFunc(): Error calling FrameData_update\n" );
		return -1;
	}

	
	// Now let's call our drawing function.
	DrawGL();
	
	return noErr;
}


	